@@ -27,7 +27,8 @@ import tabulate
 from mdutils.mdutils import MdUtils
 
 DEFAULT_INI = {'global': {'unusedholes': 'yes',
-                          'onebitenum': 'no'}}
+                          'onebitenum': 'no',
+                          'PublicConvFunct': 'no'}}
 
 
 def sortRegisterAndFillHoles(regName,
@@ -172,12 +173,13 @@ class enumTypeClass():
 class rstAddressBlock(addressBlockClass):
     """Generates a ReStructuredText file from a IP-XACT register description"""
 
-    def __init__(self, name, addrWidth, dataWidth):
+    def __init__(self, name, addrWidth, dataWidth, config):
         self.name = name
         self.addrWidth = addrWidth
         self.dataWidth = dataWidth
         self.registerList = []
         self.suffix = ".rst"
+        self.config = config
 
     def returnEnumValueString(self, enumTypeObj):
         if isinstance(enumTypeObj, enumTypeClass):
@@ -292,7 +294,7 @@ class rstAddressBlock(addressBlockClass):
 class mdAddressBlock(addressBlockClass):
     """Generates a Markdown file from a IP-XACT register description"""
 
-    def __init__(self, name, addrWidth, dataWidth):
+    def __init__(self, name, addrWidth, dataWidth, config):
         self.name = name
         self.addrWidth = addrWidth
         self.dataWidth = dataWidth
@@ -394,12 +396,13 @@ class mdAddressBlock(addressBlockClass):
 class vhdlAddressBlock(addressBlockClass):
     """Generates a vhdl file from a IP-XACT register description"""
 
-    def __init__(self, name, addrWidth, dataWidth):
+    def __init__(self, name, addrWidth, dataWidth, config):
         self.name = name
         self.addrWidth = addrWidth
         self.dataWidth = dataWidth
         self.registerList = []
         self.suffix = "_vhd_pkg.vhd"
+        self.config = config
 
     def returnAsString(self):
         r = ''
@@ -453,22 +456,14 @@ class vhdlAddressBlock(addressBlockClass):
 
         r += self.returnRegistersInRecordTypeString()
         r += self.returnRegistersOutRecordTypeString()
+        r += self.returnRegistersReadFunction()
+        r += self.returnRegistersWriteFunction()
+        r += self.returnRegistersResetFunction()
 
-        t = "  function read_" + self.name + "(registers_i : " + self.name + "_in_record_type;\n"
-        r += t
-        indent = t.find('(') + 1
-        r += " " * indent + "registers_o : " + self.name + "_out_record_type;\n"
-        r += " " * indent + "address : std_ulogic_vector(addr_width-1 downto 0)\n"
-        r += " " * indent + ") return std_ulogic_vector;\n\n"
-
-        t = "  function write_" + self.name + "(value : std_ulogic_vector(data_width-1 downto 0);\n"
-        r += t
-        indent = t.find('(') + 1
-        r += " " * indent + "address : std_ulogic_vector(addr_width-1 downto 0);\n"
-        r += " " * indent + "registers_o : " + self.name + "_out_record_type\n"
-        r += " " * indent + ") return " + self.name + "_out_record_type;\n\n"
-
-        r += "  function reset_" + self.name + " return " + self.name + "_out_record_type;\n\n"
+        if self.config['global'].getboolean('VhdlPublicConvFunct'):
+            for reg in self.registerList:
+                r += self.returnRecToSulvFunction(reg)
+                r += self.returnSulvToRecFunction(reg)
 
         r += "end;\n"
 
@@ -591,6 +586,30 @@ class vhdlAddressBlock(addressBlockClass):
         r += "  end record;\n\n"
         return r
 
+    def returnRegistersReadFunction(self):
+        r = "  function read_" + self.name + "(registers_i : " + self.name + "_in_record_type;\n"
+        indent = r.find('(') + 1
+        r += " " * indent + "registers_o : " + self.name + "_out_record_type;\n"
+        r += " " * indent + "address : std_ulogic_vector(addr_width-1 downto 0)\n"
+        r += " " * indent + ") return std_ulogic_vector;\n\n"
+        return r
+
+    def returnRegistersWriteFunction(self):
+        r = "  function write_" + self.name + "(value : std_ulogic_vector(data_width-1 downto 0);\n"
+        indent = r.find('(') + 1
+        r += " " * indent + "address : std_ulogic_vector(addr_width-1 downto 0);\n"
+        r += " " * indent + "registers_o : " + self.name + "_out_record_type\n"
+        r += " " * indent + ") return " + self.name + "_out_record_type;\n\n"
+        return r
+
+    def returnRegistersResetFunction(self):
+        r = "  function reset_" + self.name + " return " + self.name + "_out_record_type;\n"
+        r += "  function reset_" + self.name + "(address: std_ulogic_vector(addr_width-1 downto 0);\n"
+        indent = r.splitlines()[-1].find('(') + 1
+        r += " " * indent + "registers_o : " + self.name + "_out_record_type\n"
+        r += " " * indent + ") return " + self.name + "_out_record_type;\n\n"
+        return r
+
     def returnRecToSulvFunctionString(self, reg):
         r = ""
         r += "  function " + reg.name + \
@@ -617,6 +636,13 @@ class vhdlAddressBlock(addressBlockClass):
         r += "  end function;\n\n"
         return r
 
+    def returnRecToSulvFunction(self, reg):
+
+        r = ""
+        r += "  function " + reg.name + \
+             "_record_type_to_sulv(v : " + reg.name + "_record_type) return std_ulogic_vector;\n\n"
+        return r
+
     def returnSulvToRecFunctionString(self, reg):
         r = ""
         r += "  function sulv_to_" + reg.name + \
@@ -642,6 +668,12 @@ class vhdlAddressBlock(addressBlockClass):
         r += "  end function;\n\n"
 
         return r
+		
+    def returnSulvToRecFunction(self, reg):
+        r = ""
+        r += "  function sulv_to_" + reg.name + \
+             "_record_type(v : std_ulogic_vector) return " + reg.name + "_record_type;\n\n"
+        return r
 
     def returnReadFunctionString(self):
         r = ""
@@ -700,6 +732,24 @@ class vhdlAddressBlock(addressBlockClass):
                     r += "         r." + reg.name + " := sulv_to_" + \
                          reg.name + "_record_type(" + reg.name + "_reset_value);\n"
         r += "    return r;\n"
+        r += "  end function;\n"
+        r += "\n"
+        r += "  function reset_" + self.name + "(address: std_ulogic_vector(addr_width-1 downto 0);\n"
+        indent = r.splitlines()[-1].find('(') + 1
+        r += " " * indent + "registers_o : " + self.name + "_out_record_type\n"
+        r += " " * indent + ") return " + self.name + "_out_record_type is\n"
+        r += "    variable r : " + self.name + "_out_record_type;\n"
+        r += "  begin\n"
+        r += "    r := registers_o;\n"
+        r += "    case to_integer(unsigned(address)) is\n"
+        for reg in self.registerList:
+            if reg.resetValue:
+                if reg.access != "read-only":
+                    r += "         when " + reg.name + "_addr => r." + reg.name + \
+                        " := sulv_to_" + reg.name + "_record_type(" + reg.name + "_reset_value);\n"
+        r += "      when others => null;\n"
+        r += "    end case;\n"
+        r += "    return r;\n"
         r += "  end function;\n\n"
         return r
 
@@ -721,12 +771,13 @@ class vhdlAddressBlock(addressBlockClass):
 
 
 class systemVerilogAddressBlock(addressBlockClass):
-    def __init__(self, name, addrWidth, dataWidth):
+    def __init__(self, name, addrWidth, dataWidth, config):
         self.name = name
         self.addrWidth = addrWidth
         self.dataWidth = dataWidth
         self.registerList = []
         self.suffix = "_sv_pkg.sv"
+        self.config = config
 
     def returnIncludeString(self):
         r = "\n"
@@ -876,6 +927,115 @@ class systemVerilogAddressBlock(addressBlockClass):
         r += "endpackage //" + self.name + "_sv_pkg\n"
         return r
 
+        
+class cAddressBlock(addressBlockClass):
+    def __init__(self, name, addrWidth, dataWidth, config):
+        self.name = name
+        self.addrWidth = addrWidth
+        self.dataWidth = dataWidth
+        self.registerList = []
+        self.suffix = ".h"
+        
+    def registerOffsetName(self, reg):
+        return self.name.upper() +"_" + reg.name.upper() + "_OFFSET"
+        
+    def registerResetName(self, reg):
+        return self.name.upper() +"_" + reg.name.upper() + "_RESET"
+
+    def getFieldMaskName(self, reg, fieldname):
+        return self.name.upper() + "_" + reg.name.upper() + "_" + fieldname.upper() + "_MASK"
+        
+    def getFieldShiftName(self, reg, fieldname):
+        return self.name.upper() + "_" + reg.name.upper() + "_" + fieldname.upper() + "_SHIFT"
+
+    def getMacroName(self, reg, fieldname):
+        return "GET_" + self.name.upper() + "_" + reg.name.upper() + "_" + fieldname.upper()
+        
+    def returnRegisterOffsets(self):
+        r = ""
+        r += "// ------------------------------------------------ \n"
+        r += "//  Register offsets\n"
+        r += "// ------------------------------------------------ \n"
+        for reg in self.registerList:
+            addrStr = "0x%0.2X" % reg.address
+            r += "#define "+ self.registerOffsetName(reg) + "\t" + addrStr + "\t// " + reg.desc +"\n"
+        
+        r += "\n\n"
+        return r
+        
+    def returnRegisterBitOperators(self):
+        r = ""
+        
+        for reg in self.registerList:
+            r += "// ------------------------------------------------ \n"
+            r += "//  Bit operations for register " + reg.name + "\n"
+            r += "// ------------------------------------------------ \n"
+            for i in list(range(len(reg.fieldNameList))):
+                fieldname = reg.fieldNameList[i]
+                r += "#define "+ self.getFieldShiftName(reg, fieldname) + "\t" + \
+                     str(reg.bitOffsetList[i]) + "\n"
+                mask = 2 ** reg.bitWidthList[i] -1;
+                maskStr = "0x%0.2X" % mask
+                r += "#define "+ self.getFieldMaskName(reg, fieldname)  + " \t" + \
+                     maskStr + "\n"
+                     
+                r += "\n"
+        
+        return r
+        
+    def returnMacrosFunctions(self):
+        r = ""
+        
+        for reg in self.registerList:
+            r += "\n"
+            r += "// ------------------------------------------------ \n"
+            r += "//  Macro functions for register " + reg.name + "\n"
+            for i in list(range(len(reg.fieldNameList))):
+                fieldname = reg.fieldNameList[i]
+                r += "//  - " + self.getMacroName(reg, fieldname) + " : " + reg.fieldDescList[i] + "\n" 
+            r += "// ------------------------------------------------\n"
+            r += "\n"
+            for i in list(range(len(reg.fieldNameList))):
+                fieldname = reg.fieldNameList[i]
+                operation = "((a >> " + self.getFieldShiftName(reg, fieldname) + \
+                            ") & " + self.getFieldMaskName(reg, fieldname) + ")"
+                            
+                r += "#define "+ self.getMacroName(reg, fieldname) + "(a)\t" + operation + "\n"
+        
+        return r
+
+    def returnAsString(self):
+        r = ''
+        r += "#pragma once\n"
+        r += "/* Automatically generated\n"
+        r += " * with the command '%s'\n" % (' '.join(sys.argv))
+        r += " *\n"
+        r += " * Do not manually edit!\n"
+        r += " *\n"
+        r += " * Example usage:\n"
+        r += " *     uint32_t datareg0 = read(dev, EXAMPLE_REG_ADDRESS_REG0);\n"
+        r += " *\n"
+        r += " *     uint8_t byte0 = (uint8_t)GET_EXAMPLE_REG0_BYTE0(datareg0);\n"
+        r += " *     uint8_t byte1 = (uint8_t)GET_EXAMPLE_REG0_BYTE1(datareg0);\n"
+        r += " *     uint8_t byte2 = (uint8_t)GET_EXAMPLE_REG0_BYTE2(datareg0);\n"
+        r += " *     uint8_t byte3 = (uint8_t)GET_EXAMPLE_REG0_BYTE3(datareg0);\n"
+        r += " *\n"
+        r += " *\n"
+        r += " *     uint32_t datareg7 = read(dev, EXAMPLE_REG_ADDRESS_REG7);\n"
+        r += " *\n"
+        r += " *     uint8_t nibble0 = (uint8_t)GET_EXAMPLE_REG7_NIBBLE0(datareg7);\n"
+        r += " *     uint8_t nibble1 = (uint8_t)GET_EXAMPLE_REG7_NIBBLE1(datareg7);\n"
+        r += " *     uint8_t nibble2 = (uint8_t)GET_EXAMPLE_REG7_NIBBLE2(datareg7);\n"
+        r += " */\n"
+        
+        r += self.returnRegisterOffsets()
+        r += self.returnRegisterBitOperators()
+        r += self.returnMacrosFunctions()
+            
+        r += "\n"
+        r += "// End of " + self.name + self.suffix + "\n"
+        return r
+
 
 class ipxactParser():
     def __init__(self, srcFile, config):
@@ -980,9 +1140,10 @@ class ipxactParser():
 
 
 class ipxact2otherGenerator():
-    def __init__(self, destDir, namingScheme="addressBlockName"):
+    def __init__(self, destDir, config, namingScheme="addressBlockName", ):
         self.destDir = destDir
         self.namingScheme = namingScheme
+        self.config = config
 
     def write(self, fileName, string):
         _dest = os.path.join(self.destDir, fileName)
@@ -1001,9 +1162,13 @@ class ipxact2otherGenerator():
             mapName = memoryMap.name
             for addressBlock in memoryMap.addressBlockList:
                 blockName = addressBlock.name
+
                 block = generatorClass(addressBlock.name,
                                        addressBlock.addrWidth,
-                                       addressBlock.dataWidth)
+                                       addressBlock.dataWidth,
+                                       self.config,
+                                       )
+
                 block.setRegisterList(addressBlock.registerList)
                 s = block.returnAsString()
                 if self.namingScheme == "addressBlockName":
